@@ -2,4 +2,8 @@ from typing import *
 
 
 def normalize_axis_index(_0: int, _1: int, /):
-    "usage.xarray: 3"
+    """
+    usage.xarray: 3
+    usage.dask: 1
+    """
+    ...
