@@ -18,9 +18,9 @@ if __name__ == "__main__":
     try:
         with tracer:
             runpy.run_module(
-                tracer.calls_from_module, run_name="__main__", alter_sys=True
+                tracer.calls_from_modules[0], run_name="__main__", alter_sys=True
             )
     except Exception:
-        raise Exception(f"Error running {tracer.calls_from_module}")
+        raise Exception(f"Error running {tracer.calls_from_modules}")
     finally:
         finalize()
