@@ -22,7 +22,6 @@ __all__ = ["Tracer", "setup", "finalize", "get_tracer"]
 
 DEBUG = os.environ.get("PYTHON_RECORD_API_DEBUG", False)
 
-MAX_LENGTH = 50
 
 # global cache for tracer based on env variables
 TRACER = None
@@ -34,9 +33,9 @@ write_line: Optional[Callable[[dict], None]] = None
 def get_tracer() -> Tracer:
     global TRACER
     if not TRACER:
-        FROM_MODULE = os.environ["PYTHON_RECORD_API_TO_MODULE"]
-        TO_MODULE = os.environ["PYTHON_RECORD_API_FROM_MODULE"]
-        TRACER = Tracer(FROM_MODULE, TO_MODULE)
+        FROM_MODULES = os.environ["PYTHON_RECORD_API_TO_MODULES"].split(",")
+        TO_MODULES = os.environ["PYTHON_RECORD_API_FROM_MODULES"].split(",")
+        TRACER = Tracer(FROM_MODULES, TO_MODULES)
     return TRACER
 
 
@@ -71,6 +70,23 @@ def _encode_slice(s: slice):
     }
 
 
+# # cache this b/c its expesnive
+# @functools.lru_cache(None, False)
+# def find_root_module(o: object):
+#     """
+#     Looks from the top of the module tree down to see if it is exportd from a parent module
+#     """
+#     full_module = o.__module__
+#     modules = full_module.split(".")
+#     if len(modules) == 1:
+#         return modules[0]
+#     for i in range(len(modules) - 1):
+#         parent_module = ".".join(modules[i:])
+#         mod = importlib.import_module(parent_module)
+#         # Export by module
+#         if o in mod
+
+
 @encode.register(types.FunctionType)
 @encode.register(type)
 def encode_module_value(v):
@@ -102,6 +118,16 @@ def _encode_method(m: types.MethodType):
 
 @encode.register
 def _encode_method_descriptor(m: types.MethodDescriptorType):
+    """
+    This is actually a method, not a classmethod!
+
+    np.ndarray.mean(x)
+    or
+    np.arange(10).mean()
+
+    both end up being the same call, we can't tell the difference.
+    """
+    # TODO: Remove class, we can grab off of first arg
     return {"name": m.__name__, "class": m.__objclass__}
 
 
@@ -128,8 +154,10 @@ else:
         return u._func.__name__
 
 
-MAX_SEQUENCE = 5
-MAX_STRING = 50
+MAX_LIST = 10
+MAX_TUPLE = 10
+MAX_DICT = 10
+MAX_STRING = 20
 
 
 def preprocess(o):
@@ -141,21 +169,21 @@ def preprocess(o):
     All non primitives subclasses are passed through to be handled by `default`.
     """
     tp = type(o)
-    if tp == str:
-        return o[:MAX_STRING]
+    if tp == str and len(o) <= MAX_STRING:
+        return o
     if tp == list:
-        return [preprocess(v) for v in o[:MAX_SEQUENCE]]
+        return [preprocess(v) for v in o[:MAX_LIST]]
     if tp == dict:
         return {
             "t": "dict",
             "v": [
-                [preprocess(k), preprocess(v)] for k, v in list(o.items())[:MAX_LENGTH]
+                [preprocess(k), preprocess(v)] for k, v in list(o.items())[:MAX_DICT]
             ],
         }
     if tp == tuple:
         return {
             "t": "tuple",
-            "v": [preprocess(v) for v in o[:MAX_SEQUENCE]],
+            "v": [preprocess(v) for v in o[:MAX_TUPLE]],
         }
     if isinstance(o, (int, float, bool, list, str, dict, tuple)):
         # Don't send literals of these
@@ -176,7 +204,6 @@ def default(o: object) -> object:
     return {"t": t}
 
 
-
 # cache this b/c its expesnive
 @functools.lru_cache(None, False)
 def signature(fn):
@@ -230,6 +257,7 @@ class Bound:
         return b
 
 
+# TODO: Make not args kwargs, since kwarg could be fn
 def log_call(location: str, fn: Callable, *args, **kwargs) -> None:
     bound = Bound.create(fn, args, kwargs)
     line: Dict = {"location": location, "function": preprocess(fn)}
@@ -405,7 +433,7 @@ class Stack:
         }
         if val in COMPARISONS:
             self.process(
-                (self.TOS, self.TOS1),
+                (self.TOS, self.TOS1) if "in" not in val else (self.TOS),
                 COMPARISONS[val],
                 # "in" and "not in" are reversed
                 (self.TOS1, self.TOS,) if "in" not in val else (self.TOS, self.TOS1,),
@@ -500,10 +528,10 @@ BINARY_OPS = {
 
 @dataclasses.dataclass
 class Tracer:
-    # the module we should trace calls to
-    calls_to_module: str
-    # the module we should trace calls from
-    calls_from_module: str
+    # the modules we should trace calls to
+    calls_to_modules: List[str]
+    # the modules we should trace calls from
+    calls_from_modules: List[str]
 
     def __enter__(self):
         sys.settrace(self)
@@ -521,7 +549,7 @@ class Tracer:
             if not module:
                 warnings.warn(f"Cannot get module of {value}")
                 continue
-            if module.startswith(self.calls_to_module):
+            if any(module.startswith(mod) for mod in self.calls_to_modules):
                 return True
         return False
 
@@ -542,10 +570,9 @@ class Tracer:
             frame_module_name = frame.f_globals["__name__"]
         except KeyError:
             return False
-        return (
-            frame_module_name == "__main__"
-            or frame_module_name == self.calls_from_module
-            or frame_module_name.startswith(self.calls_from_module + ".")
+        return frame_module_name == "__main__" or any(
+            frame_module_name == mod or frame_module_name.startswith(mod + ".")
+            for mod in self.calls_from_modules
         )
 
 
